package emulator.hardware.nmos6502.commands;

import emulator.EmulatorException;
import emulator.hardware.HwByte;
import emulator.hardware.nmos6502.Command;
import emulator.hardware.nmos6502.CommandSet;
import emulator.hardware.nmos6502.Cpu6502;
import emulator.hardware.nmos6502.Flags6502;
import emulator.hardware.nmos6502.Operand;
import emulator.hardware.nmos6502.operands.*;

/** EOR (bitwise Exclusive OR)
 *  
 *  Affects Flags: S Z
 *  
 *  MODE           SYNTAX       HEX LEN TIM
 *  Immediate     EOR #$44      $49  2   2
 *  Zero Page     EOR $44       $45  2   3
 *  Zero Page,X   EOR $44,X     $55  2   4
 *  Absolute      EOR $4400     $4D  3   4
 *  Absolute,X    EOR $4400,X   $5D  3   4+
 *  Absolute,Y    EOR $4400,Y   $59  3   4+
 *  Indirect,X    EOR ($44,X)   $41  2   6
 *  Indirect,Y    EOR ($44),Y   $51  2   5+
 *  
 *  + add 1 cycle if page boundary crossed
 */
public class EOR implements Command {

	@Override
	public void execute(Cpu6502 cpu, Operand operand) throws EmulatorException {
		HwByte value = cpu.getAccu();
		value.xor(operand.getByte());
		cpu.setAccu(value);
		Flags6502 flags = cpu.getFlags();
		flags.setNZFromValue(value);
		cpu.setFlags(flags);
	}

	@Override
	public String getName() {
		return "EOR";
	}

	@Override
	public void register(CommandSet cset) {
		cset.defineCommand(0x49, this, new Immediate());
		cset.defineCommand(0x45, this, new ZeroPage());
		cset.defineCommand(0x55, this, new ZeroPageX());
		cset.defineCommand(0x4D, this, new Absolute());
		cset.defineCommand(0x5D, this, new AbsoluteX());
		cset.defineCommand(0x59, this, new AbsoluteY());
		cset.defineCommand(0x41, this, new IndirectX());
		cset.defineCommand(0x51, this, new IndirectY());
	}

}
